import re
from functools import cached_property, cache
from pathlib import Path

from cx_core.text import is_quoted, contains_invisible_char, quote_text


class PathDataSource:
    SPACE_MODES = ['quote', 'escape', None]

    def __init__(self, path: Path, space_mode=None):
        self._source = Path(path)
        self._space_mode = space_mode

    @cached_property
    def _neat_source(self):
        return self._source.resolve()

    @cache
    def _handle(self, argument):
        match argument:
            case 'absolute':
                return self._neat_source.absolute()
            case 'name':
                return self._source.name
            case 'basename':
                return self._source.stem
            case 'suffix':
                return self._source.suffix
            case 'complete_suffix':
                return ''.join(self._source.suffixes)
            case 'suffix_no_dot':
                suffix = self._source.suffix
                if suffix.startswith('.'):
                    return suffix[1:]
                return suffix
            case 'complete_suffix_no_dot':
                suffix = ''.join(self._source.suffixes)
                if suffix.startswith('.'):
                    return suffix[1:]
                return suffix
            case 'complete_basename':
                basename = self._source.stem
                suffixes = self._source.suffixes
                return basename + ''.join(suffixes[:-1])
            case 'parent':
                return self._source.parent
            case 'parent_absolute':
                return self._neat_source.parent.absolute()
            case 'parent_name':
                return self._source.parent.name
            case _:
                return self._source

    def __call__(self, argument):
        result = str(self._handle(argument))
        if is_quoted(result) and contains_invisible_char(result):
            if self._space_mode == 'quote':
                result = quote_text(result)
            elif self._space_mode == 'escape':
                result = re.sub(r'\s+', '\\ ', result)
        return result



import glob
import pkgutil
from argparse import ArgumentParser

from rich.markdown import Markdown
from rich.panel import Panel
from rich_argparse import RichHelpFormatter

from cx_core.app import AbstractApp, LogLevel
from cx_core.filesystem import PathExpander
from .worker import *


class SubConvApp(AbstractApp):
    APP_VERSION = "0.3.7"
    APP_NAME = "subconv"

    def __init__(self):
        super(SubConvApp, self).__init__()

        _parser = ArgumentParser(prog=SubConvApp.APP_NAME,
                                 description=f'批量提取台词本，支持的格式包括{" ".join(Worker.acceptable_suffixes())}',
                                 epilog=f'Version {SubConvApp.APP_VERSION} Designed by xiii_1991',
                                 formatter_class=RichHelpFormatter, exit_on_error=False)

        _parser.add_argument('--format', '-f', dest='format', default='txt',
                             choices=Worker.supported_formats(),
                             help='指定输出格式', metavar='格式代码')

        _parser.add_argument('--appending', '-a', dest='appending', default=None,
                             help='为目标文件添加后缀', metavar='后缀字符串')

        _parser.add_argument('--encoding', '-c', dest='target_encoding', default='auto',
                             help='设定输出编码', metavar='目标编码')

        _parser.add_argument('--force-decoding', dest='source_encoding', default=None,
                             help='强制指定读取编码', metavar='来源编码')

        _parser.add_argument('-o', '--output-dir', dest='target_dir', default=None,
                             help='指定输出目录', metavar='目录')

        _parser.add_argument('--sub-dir', dest='sub_dir', default=None,
                             help='指定子目录', metavar='子目录')

        _parser.add_argument('-t', '--time', dest='keep_time', action='store_true',
                             help='尽量保留时间信息')

        _parser.add_argument('--bypass-formatter', dest='bypass_formatter', default=False,
                             action='store_true', help='跳过默认的内容检查器，使用原样输出')

        _parser.add_argument('--overwrite', '-y', dest='overwrite_target', default=False,
                             action='store_true', help='强制覆盖已存在的目标文件')

        _parser.add_argument('--translate', dest='translate', action='store_true',
                             help='自动翻译为英文版本（需要网络）')

        _parser.add_argument('--debug', '-d', action='store_true', dest='debug',
                             help='调试模式')

        _parser.add_argument('--pretend', '-p', dest='pretend_mode', action='store_true',
                             help='干转模式，不执行写入操作')

        _parser.add_argument('--full-help', help='显示详细的说明',
                             dest='show_full_help', action='store_true')

        _parser.add_argument('sources', nargs='*',
                             help='需要转换的文件', metavar='文件列表')

        self.parser = _parser
        self.profile = None

        self.global_task = None

    def __enter__(self):
        env.start()

        env.print(f'[yellow]{SubConvApp.APP_NAME}[/yellow] [blue]{SubConvApp.APP_VERSION}[/blue]')

        env.info('开始解析参数')
        _args = self.parser.parse_args()
        self.profile = DataPackage(**vars(_args))
        env.debug('解析的参数:', self.profile)

        env.log_level = LogLevel.DEBUG if self.profile.debug else LogLevel.WARNING

        self.global_task = env.progress.add_task(description='全局进度', total=None)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        result = False

        if exc_type is None:
            env.print('执行完毕，没有发现错误')
        elif issubclass(exc_type, UserCancelError):
            env.error(f'已取消未完成的任务')
            result = True

        env.progress.remove_task(self.global_task)
        self.global_task = None

        env.stop()
        return result

    @staticmethod
    def show_full_help():
        data = pkgutil.get_data('sub_conv', 'help.md').decode('utf-8')
        panel = Panel(Markdown(data), width=80)
        env.console.print(panel)

    @property
    def _preprocessed_sources(self):
        for source in self.profile.sources:
            yield from glob.iglob(source)

    def run(self):
        if self.profile.show_full_help:
            self.show_full_help()
            return

        def _validator(p: Path) -> bool:
            sub_dir = self.profile.sub_dir
            suffixes = Worker.acceptable_suffixes()
            if sub_dir and p.parent.name == sub_dir:
                return False
            return p.suffix in suffixes

        expander = PathExpander(self._preprocessed_sources,
                                accept_dir=False,
                                file_validator=_validator)
        env.debug('Expander 已构建')

        env.progress.start_task(self.global_task)
        for source in env.progress.track(expander, task_id=self.global_task):
            if env.wanna_quit:
                raise UserCancelError()
            mission = Mission(source, self.profile)
            env.debug('构建 Mission:', mission)
            env.progress.update(self.global_task, description=mission.name)
            with Worker(mission) as worker:
                worker.process()


def run():
    with SubConvApp() as app:
        app.run()



import re


class TagReplacer:
    TAG_PATTERN = re.compile(r"\${([^:{}]+)(?::([^:{}]+))?}")

    def __init__(self, keep_unknown_tags=True, default_data=""):
        self._data_sources = {}
        self._keep_unknown_tags = keep_unknown_tags
        self._default_data = default_data

    def install_data_source(self, tag: str, source):
        if callable(source):
            self._data_sources[tag] = source
        else:
            self._data_sources[tag] = str(source)

        return self

    def tags(self):
        return self._data_sources.keys()

    def __lookup_match(self, match: re.Match):
        tag = match.group(1)
        param = match.group(2) or None
        if tag in self._data_sources:
            source = self._data_sources[tag]
            if callable(source):
                return source(param)
            elif isinstance(source, str):
                return source
            else:
                return tag

        if self._keep_unknown_tags:
            return match.group(0)

        return self._default_data

    def lookup(self, tag: str):
        match = TagReplacer.TAG_PATTERN.match(tag)
        if match:
            return self.__lookup_match(match)
        return tag

    def replace_tags(self, text: str):
        return TagReplacer.TAG_PATTERN.sub(self.__lookup_match, text)

    def __call__(self, text: str):
        t = str(text or "")
        result = self.replace_tags(t)
        return result



from ffmpeg import Progress
from ffmpeg.asyncio import FFmpeg
from ffmpeg.errors import FFmpegError
from pathlib import Path
import json, asyncio
from cx_core.filesystem.path_utils import is_file_in_dir
from cx_core import DataPackage
from .env import env
from .mission_maker import Mission


class Transcoder:
    def __init__(self, mission: Mission) -> None:
        self.task = None
        self.mission = mission
        self.is_valid = True
        self.media_info = None
        self.ffmpeg = None
        self.duration = None

    def __enter__(self):
        self.task = env.progress.add_task(
            description=self.mission.source.name, visible=False, total=None
        )

        for o in self.mission.outputs:
            f = o.filename
            if f.exists() and not self.mission.overwrite:
                env.warning(f"目标文件[red]{f}[/red]已存在且不可覆盖，该任务无法执行")
                self.is_valid = False
                break

        for i in self.mission.inputs:
            if not i.filename.exists():
                env.warning(f"文件[red]{i.filename}[/red]不存在，该任务无法执行")
                self.is_valid = False
                break
            if is_file_in_dir(i.filename, self.mission.target_folder):
                env.warning(
                    f"文件[red]{i.filename}[/red]存在于目标目录中，此任务无法执行"
                )
                self.is_valid = False
                break

        ffprobe_bin = Path(str(self.mission.ffmpeg).replace("ffmpeg", "ffprobe"))
        probe_file = self.mission.source
        if not probe_file.exists():
            probe_file = None
            for f in self.mission.inputs:
                if f.filename.exists():
                    probe_file = f.filename
                    break

        if not probe_file:
            env.warning(
                f"从源文件[red]{self.mission.source.name}[/red]获取媒体元数据失败"
            )
        else:
            ffprobe = FFmpeg(executable=ffprobe_bin).input(
                probe_file, print_format="json", show_format=None
            )
            media = None
            try:
                media = json.loads(asyncio.run(ffprobe.execute()))
            finally:
                if (media is not None) and any(media):
                    self.media_info = DataPackage(**media)
                else:
                    env.warning(
                        f"无法为[cyan]{self.mission.source.name}[/cyan]读取元数据，无法进行时间估计"
                    )

        if self.media_info:
            self.duration = float(self.media_info.format.duration)
            env.progress.update(self.task, total=float(self.duration))

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        result = False
        if exc_type is None:
            pass
        elif issubclass(exc_type, FFmpegError):
            env.critical(f"FFMPEG执行出错：[red]{exc_val}[/red]")
            result = True
        env.progress.stop_task(self.task)
        env.progress.remove_task(self.task)
        return result

    async def transcode(self):
        if env.wanna_quit:
            env.debug(f"检测到用户中断，直接跳过任务")
            return

        if not self.is_valid:
            env.debug(f"任务非法，跳过执行")
            return

        env.progress.update(
            self.task, description="检查目标目录可用性…", completed=0.0, visible=True
        )
        env.progress.start_task(self.task)
        for o in self.mission.outputs:
            f = o.filename.parent
            if not f.exists():
                f.mkdir(parents=True, exist_ok=True)
                env.debug(f"创建文件夹{f}")

        self.ffmpeg = FFmpeg(self.mission.ffmpeg)
        env.debug(f"创建ffmpeg对象 {self.ffmpeg} ({self.mission.ffmpeg})")

        env.progress.update(self.task, description="写入全局参数…")
        for k, v in self.mission.general.iter_options():
            self.ffmpeg.option(k, v)
            env.debug(f"写入全局选项: [blue]{k}[/blue] : [yellow]{v}[/yellow]")

        for i in env.progress.track(
            self.mission.inputs, task_id=self.task, description="写入输入参数…"
        ):
            self.ffmpeg.input(i.filename.absolute(), i.data)
            env.debug(f"为[cyan]{i.filename}[/cyan]添加输入选项：{i.data}")

        for o in env.progress.track(
            self.mission.outputs, task_id=self.task, description="写入输出参数…"
        ):
            self.ffmpeg.output(o.filename.absolute(), o.data)
            env.debug(f"为[cyan]{o.filename}[/cyan]添加输出选项：{o.data}")

        env.progress.update(self.task, description="设置事件监听器…")

        @self.ffmpeg.on("progress")
        def on_progress(progress: Progress):
            desc = f"{self.mission.source.name} [yellow]x{progress.speed:.2f}[/yellow] [green]{progress.bitrate:.2f}k/s[/green]"
            current = progress.time.total_seconds()
            env.progress.update(
                self.task,
                description=desc,
                completed=float(current),
                total=self.duration,
            )
            if env.wanna_quit:
                self.ffmpeg.terminate()

        @self.ffmpeg.on("stderr")
        def on_stderr(line):
            env.debug(f"[grey]FFMPEG标准错误：[/grey]{line}")

        @self.ffmpeg.on("start")
        def on_start(arguments):
            env.debug(f"FFMPEG启动：[grey]{" ".join(arguments)}[/grey]")

        @self.ffmpeg.on("completed")
        def on_completed():
            env.debug(f"[green]{self.mission.source.name}[/green]执行完毕")
            env.progress.update(self.task, description="转码完成")

        @self.ffmpeg.on("terminated")
        def on_terminated():
            env.warning("[purple]FFMPEG被终止[/purple]")
            env.progress.update(self.task, description="尝试移除未完成的目标文件…")
            for t in self.mission.outputs:
                f = t.filename
                if f.exists:
                    f.unlink(missing_ok=True)
                    env.print(f"移除未完成的目标文件[red]{f}[/red]")

        try:
            env.progress.update(self.task, description="正在启动…")
            await self.ffmpeg.execute()
        except FFmpegError as e:
            env.error(f"FFMPEG运行异常：[red]{e.message}[/red]")
            self.ffmpeg.terminate()

    def run(self):
        asyncio.run(self.transcode())



from chardet import UniversalDetector

__char_detector = UniversalDetector()


def detect_encoding(filename):
    __char_detector.reset()
    try:
        with open(filename, 'rb') as fp:
            for line in fp.readlines():
                __char_detector.feed(line)
                if __char_detector.done:
                    break
            result = __char_detector.result
            return result['encoding']
    except FileNotFoundError:
        return 'utf-8'



from cx_core.app import CxException


class NoProfileError(CxException):

    def __init__(self, message=None):
        self.message = message
        super().__init__(self.message)


class NoSourcesError(CxException):
    def __init__(self, message=None):
        self.message = message
        super().__init__(self.message)


class UserCanceledError(CxException):
    def __init__(self, message=None):
        self.message = message
        super().__init__(self.message)



from pathlib import Path
from .env import env
from cx_core.misc import DataPackage
import tomllib


class ProfileLoader:
    def __init__(self, filename: Path, args: DataPackage = None) -> None:
        self.task = None
        self.filename = filename
        self.args = args

    def __enter__(self):
        self.task = env.progress.add_task(
            "解析配置：[yellow]{self.filename}[/yellow]",
            start=False,
            visible=True,
            total=None,
        )
        return self

    def __exit__(self, a, b, c):
        env.progress.stop_task(self.task)
        env.progress.remove_task(self.task)
        return False

    def load(self) -> DataPackage:
        data = {}
        env.progress.start_task(self.task)
        with env.progress.open(self.filename, "rb", task_id=self.task) as f:
            data.update(tomllib.load(f))
        data["path"] = Path(self.filename).resolve()
        env.debug("toml解析结果", data)

        result = DataPackage(**data)
        env.print(
            f"发现配置文件 <[cyan]{result.general.name}[/cyan]> [yellow]{result.general.description}[/yellow]"
        )

        if self.args and self.args.output_dir:
            result.target.folder = (
                Path(self.args.output_dir) / Path(result.target.folder).name
            ).resolve()
            env.print(
                f"因指定了输出目录，配置文件的输出位置将被设置为: [green]{result.target.folder}[/green]"
            )

        if not result.general.ffmpeg:
            result.general.ffmpeg = "ffmpeg"
            env.print("使用系统环境中的 ffmpeg")

        if result.general.overwrite is None:
            result.general.overwrite = False
            env.print("未指定覆写模式，默认关闭")

        if not result.target.folder:
            result.target.folder = "."
            env.print("未指定目标目录，将使用当前工作目录")

        if not result.target.keep_parent_level:
            result.target.keep_parent_level = 0
            env.print("未指定目标父目录保留层级，默认不保留")

        result.path = Path(self.filename)

        env.debug(f"参数包解析结果：{result}")
        return result



from cx_core.misc import DataPackage
from pathlib import Path
from cx_core.text import TagReplacer
from dataclasses import dataclass
from functools import cache
from cx_core.text.text_utils import unquote_text, split_at_unquoted_spaces
from cx_core.filesystem import normalize_path
from cx_core.misc.misc_utils import limit_number
from collections import defaultdict
from .env import env


class OptionPackage:
    def __init__(self, filename=None):
        self.data = defaultdict(list)
        self.filename = Path(filename) if filename else None

    @staticmethod
    def __format_key(k: str):
        key = str(k).strip()
        if key.startswith("-"):
            key = key[1:]
        return key

    @staticmethod
    def __format_value(v):
        value = str(v) if v else ""
        return unquote_text(value)

    def insert(self, key, value=None):
        k = OptionPackage.__format_key(key)
        v = OptionPackage.__format_value(value) if value else None
        if v:
            self.data[k].append(v)
        elif k not in self.data:
            self.data[k] = []
        return self

    def iter_options(self):
        for k, vs in self.data.items():
            if not vs:
                yield k, None
            else:
                for v in vs:
                    yield k, v

    def iter_arguments(self):
        for k, v in self.iter_options():
            yield "-" + k
            yield v

    def __rich_repr__(self):
        yield "filename", self.filename
        yield "options", self.data


@dataclass
class Mission:
    source: Path = None
    ffmpeg: str = None
    overwrite: bool = None
    target_folder: Path = None
    general: OptionPackage = None
    inputs: list[OptionPackage] = None
    outputs: list[OptionPackage] = None

    def __rich_repr__(self):
        yield "Mission"
        yield "Source", self.source
        yield "target_folder", self.target_folder
        yield "general", self.general.iter_arguments()
        yield "inputs", self.inputs
        yield "outputs", self.outputs

    def iter_output_filenames(self):
        for x in self.outputs:
            yield x.filename


class MissionMaker:
    class OptionParser:
        def __init__(self, options):
            self._options = options

        @staticmethod
        def _iter_list(elements: list):
            prev_token = None
            for t in elements:
                token = str(t).strip()
                if token.startswith("-"):
                    if prev_token:
                        yield prev_token, None
                        prev_token = None
                    else:
                        prev_token = token[1:]
                else:
                    if prev_token:
                        yield prev_token, token
                        prev_token = None
                    else:
                        env.debug(f"忽略无法识别的参数： {token}")
            if prev_token:
                yield prev_token, None

        @staticmethod
        def _iter_dict(elements: dict):
            for k, v in elements.items():
                yield k, v

        def __iter__(self):
            if isinstance(self._options, list):
                return self._iter_list(self._options)
            if isinstance(self._options, dict | DataPackage):
                return self._iter_dict(self._options)
            return self._iter_list(split_at_unquoted_spaces(str(self._options)))

    def __init__(self, source: Path, profile: DataPackage) -> None:
        self.source = Path(source)
        self.profile = profile
        self.replacer = TagReplacer(keep_unknown_tags=False)
        self.replacer.install_data_source(
            "source", self.__datahandler_source
        ).install_data_source("target", self.__datahandler_target).install_data_source(
            "source_parent", self.__datahandler_sourceparent
        ).install_data_source(
            "target_parent", self.__datahandler_targetparent
        ).install_data_source(
            "profile", self.__datahandler_profile
        ).install_data_source(
            "custom", self.__datahandler_custom
        )

    @property
    @cache
    def target(self) -> Path:
        profile_target_folder = self.replacer(self.profile.target.folder)
        target_folder = normalize_path(profile_target_folder)

        t_suffix = self.profile.target.suffix
        if not t_suffix.startswith("."):
            t_suffix = "." + t_suffix

        p_folder = Path()
        p_level = self.profile.target.keep_parent_level
        if p_level > 0:
            parents = self.source.parent.parts
            selected_parts = parents[-1 * p_level :]
            p_folder = Path(*selected_parts)
            env.debug(f"取用 {p_level} 个上级目录：{p_folder}")

        result = target_folder / p_folder / self.source.name
        return result.with_suffix(t_suffix)

    @cache
    def __datahandler_source(self, param=None):
        match param:
            case "absolute":
                return str(self.source.absolute())
            case "dot_suffix":
                return str(self.source.suffix)
            case "suffix":
                return str(self.source.suffix)[1:]
            case "parent":
                return str(self.source.parent)
            case "parent_name":
                return str(self.source.parent.stem)
            case "name":
                return str(self.source.name)
            case "basename":
                return str(self.source.stem)
            case _:
                return str(self.source)

    @cache
    def __datahandler_target(self, param=None):
        match param:
            case "absolute":
                return str(self.target.absolute())
            case "dot_suffix":
                return str(self.target.suffix)
            case "suffix":
                return str(self.target.suffix)[1:]
            case "parent":
                return str(self.target.parent)
            case "parent_name":
                return str(self.target.parent.stem)
            case "name":
                return str(self.target.name)
            case "basename":
                return str(self.target.stem)
            case _:
                return str(self.target)

    @cache
    def __datahandler_sourceparent(self, param=None):
        level = int(param) if param else 1
        ps = self.source.parent.parts
        level = limit_number(level, 1, len(ps))
        selected_parts = ps[-1 * level :]
        return str(Path(*selected_parts))

    @cache
    def __datahandler_targetparent(self, param=None):
        level = int(param) if param else 1
        ps = self.target.parent.parts
        level = limit_number(level, 1, len(ps))
        selected_parts = ps[-1 * level :]
        return str(Path(*selected_parts))

    @cache
    def __datahandler_profile(self, param=None):
        match param:
            case "id":
                return str(self.profile.general.profile_id)
            case "name":
                return str(self.profile.general.name)
            case "desc":
                return str(self.profile.general.description)
            case "ffmpeg":
                return str(self.profile.general.ffmpeg)
            case _:
                return str(self.profile.path.basename)

    @cache
    def __datahandler_custom(self, param=None):
        if param:
            return self.profile.get(f"custom.{str(param)}", str(param))
        return str(param)

    def __parse_general_table(self, general_table: DataPackage):
        result = OptionPackage()
        option_pairs = MissionMaker.OptionParser(general_table.options)
        for k, v in option_pairs:
            value = self.replacer(v)
            result.insert(k, value)

        if general_table.hardware_accelerate:
            result.insert("hwaccel", general_table.hardware_accelerate)

        overwrite_option = "-y" if general_table.overwrite else "-n"
        result.insert(overwrite_option)
        return result

    def __parse_io_table(self, io_table: DataPackage) -> OptionPackage:
        filename = self.replacer(io_table.filename)
        result = OptionPackage(filename=filename)
        opt_parser = MissionMaker.OptionParser(io_table.options)
        for k, v in opt_parser:
            result.insert(k, self.replacer(v))
        return result

    def make(self) -> Mission:
        return Mission(
            source=self.source,
            ffmpeg=str(self.profile.general.ffmpeg),
            overwrite=self.profile.general.overwrite,
            target_folder=Path(self.profile.target.folder),
            general=self.__parse_general_table(self.profile.general),
            inputs=[self.__parse_io_table(i) for i in self.profile.input],
            outputs=[self.__parse_io_table(o) for o in self.profile.output],
        )
        # result.source = self.source
        # result.ffmpeg = str(self.profile.general.ffmpeg)
        # result.overwrite = self.profile.general.overwrite
        # result.target_folder = Path(self.profile.target.folder)
        # result.general = self.__parse_general_table(self.profile.general)
        # result.inputs = [self.__parse_io_table(i) for i in self.profile.input]
        # result.outputs = [self.__parse_io_table(o) for o in self.profile.output]
        # return result



import re


def split_at_unquoted_spaces(line: str):
    result = []
    buffer = []
    symbols = {'"', "'", '(', ')', '[', ']', '{', '}'}
    matching = {')': '(', ']': '[', '}': '{'}
    stack = []
    in_quotes = False
    quote_char = ''

    for char in str(line):
        if char in symbols:
            if in_quotes:
                if char == quote_char:
                    in_quotes = False
            else:
                if char == '"' or char == "'":
                    in_quotes = True
                    quote_char = char
                elif char in matching and stack and stack[-1] == matching[char]:
                    stack.pop()
                else:
                    stack.append(char)
        if char == ' ' and not stack and not in_quotes:
            result.append(''.join(buffer))
            buffer = []
        else:
            buffer.append(char)

    # 如果缓冲区还有字符，则将它们加入到结果列表中。
    if buffer:
        result.append(''.join(buffer))

    return result


def unquote_text(s: str):
    """删除字符串前后的成对引号以及多余的空格"""
    pattern = r'^("|\')((?:(?=(\\?))\3.)*?)\1$'
    text = s.strip()
    match = re.match(pattern, text)
    if match:
        text = match.group(2)
    return text.strip()


def quote_text(s: str, quote='"'):
    """阔起字符串，默认使用引号。如果给定的quote包含多个字符，则会使用前两个分别阔起，例如使用括号"""
    left = quote[0]
    right = quote[1] if len(quote) > 1 else left
    return left + str(s) + right


def is_quoted(s: str, quotes='"\''):
    """判断字符串是否包含引号"""
    s = s.strip()
    for q in quotes:
        if s.startswith(q) and s.endswith(q):
            return True
    return False


def contains_invisible_char(s: str):
    regex = re.compile(r'\s')
    return regex.search(s)


def extract_numbers(input_string) -> [float]:
    if not input_string:
        return []
    # 使用正则表达式匹配字符串中的所有数字，包括小数
    numbers = re.findall(r'-?\d+\.?\d*', input_string)
    # 将匹配到的数字从字符串转换为浮点数或整数
    numbers = [float(num) if '.' in num else int(num) for num in numbers]
    return numbers



from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
from rich.table import Column

from .abstract_env import AbstractEnv
from .app_logger import AppLogger, LogLevel


class ProgressEnv(AbstractEnv):
    def __init__(self):
        super(ProgressEnv, self).__init__()
        self.wanna_quit = False

        self._progress = Progress(
            SpinnerColumn(),
            TextColumn('[progress.description]{task.description}', table_column=Column(ratio=50)),
            TimeElapsedColumn(),
            expand=True,
            transient=True)

        self._logger = AppLogger(console=self._progress.console)
        self._logger.level = LogLevel.WARNING

    def start(self):
        self._progress.start()
        self.info('全局环境已启动')

    def stop(self):
        self._progress.stop()
        self.info('Bye~')

    @property
    def console(self):
        return self._progress.console

    @property
    def progress(self):
        return self._progress



from collections.abc import Callable, Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from cx_core.filesystem.path_utils import normalize_path


class PathExpander:
    @dataclass
    class Settings:
        anchor_dir: Path = Path(".")
        search_subdir: bool = True
        resolve_symlinks: bool = True
        existed_only: bool = True
        accept_file: bool = True
        accept_dir: bool = True
        accept_other: bool = False
        global_validator: Callable[[Path], bool] = None
        file_validator: Callable[[Path], bool] = None
        dir_validator: Callable[[Path], bool] = None
        skip_duplicated: bool = True
        skip_unknown_type: bool = True

    @staticmethod
    def _prepare_path_list(source) -> list[Path]:
        if not isinstance(source, Iterable):
            return [Path(str(source))]
        return sorted([Path(str(x)) for x in source], key=lambda a: a.absolute())

    def __init__(
        self,
        sources: Any,
        settings: Settings = None,
        anchor_dir: Path = None,
        search_subdir: bool = None,
        resolve_symlinks: bool = None,
        existed_only: bool = None,
        accept_file: bool = None,
        accept_dir: bool = None,
        accept_other: bool = None,
        global_validator: Callable[[Path], bool] = None,
        file_validator: Callable[[Path], bool] = None,
        dir_validator: Callable[[Path], bool] = None,
        skip_duplicated: bool = None,
        **kwargs
    ):
        self._sources = PathExpander._prepare_path_list(sources)

        self.settings = settings or PathExpander.Settings()
        if anchor_dir is not None:
            self.settings.anchor_dir = Path(anchor_dir)
        if search_subdir is not None:
            self.settings.search_subdir = search_subdir
        if resolve_symlinks is not None:
            self.settings.resolve_symlinks = resolve_symlinks
        if existed_only is not None:
            self.settings.existed_only = existed_only
        if accept_file is not None:
            self.settings.accept_file = accept_file
        if accept_dir is not None:
            self.settings.accept_dir = accept_dir
        if accept_other is not None:
            self.settings.accept_other = accept_other
        if global_validator is not None:
            self.settings.global_validator = global_validator
        if file_validator is not None:
            self.settings.file_validator = file_validator
        if dir_validator is not None:
            self.settings.dir_validator = dir_validator
        if skip_duplicated is not None:
            self.settings.skip_duplicated = skip_duplicated

        if kwargs:
            self.__dict__.update(kwargs)

        self.__entry_cache = set() if self.settings.skip_duplicated else None

    def _prepare_path(self, s):
        path = normalize_path(s, self.settings.anchor_dir)
        if path.is_symlink() and self.settings.resolve_symlinks:
            path = path.resolve(strict=False)
        return path

    def _cache_entry(self, entry: Path):
        if self.__entry_cache:
            self.__entry_cache.add(entry)

    def _is_cached(self, entry: Path):
        return self.__entry_cache and (entry in self.__entry_cache)

    def clear_cache(self):
        if self.__entry_cache:
            self.__entry_cache.clear()

    def _is_acceptable_file(self, entry: Path) -> bool:
        """
        根据设置判断entry是否是一个可接受的文件。
        如果它不是文件或设置了不接受文件，则会返回 False。
        :param entry:
        :return:
        """
        if not entry.is_file():
            return False
        if not self.settings.accept_file:
            return False
        if (not entry.exists()) and self.settings.existed_only:
            return False
        if self.settings.file_validator:
            return self.settings.file_validator(entry)
        return True

    def _is_acceptable_dir(self, entry: Path) -> bool:
        """
        判断是否是一个可以接受的目录
        accept_dir 属性指定是否提出目录，并决定是否迭代它，所以此方法中不包含该判定。
        :param entry:
        :return:
        """
        if not entry.is_dir():
            return False
        if (not entry.exists()) and self.settings.existed_only:
            return False
        if self.settings.dir_validator:
            return self.settings.dir_validator(entry)
        return True

    def _expand(self, *entries):
        for entry in entries:
            entry = self._prepare_path(entry)
            if self._is_cached(entry):
                continue
            if self._is_acceptable_file(entry):
                self._cache_entry(entry)
                yield entry
            elif self._is_acceptable_dir(entry):
                if self.settings.accept_dir:
                    # 如果接受目录的话，将会迭代出目录的路径
                    yield entry
                if self.settings.search_subdir:
                    for sub_entry in entry.iterdir():
                        yield from self._expand(entry / sub_entry)
            elif self.settings.accept_other:
                self._cache_entry(entry)
                yield entry

    def __iter__(self):
        return self._expand(*self._sources)


class SuffixValidator:
    """
    简单的扩展名判断器，
    通过输入的扩展名判断路径是否合法。
    可以方便地插入到 PathExpander 中。
    """

    def __init__(self, acceptable_suffixes: Iterable):
        self._suffixes = {
            x if x.startswith(".") else "." + x
            for x in [str(a).strip().lower() for a in acceptable_suffixes]
        }

    def __call__(self, entry: Path):
        entry = Path(entry)
        return entry.suffix.lower() in self._suffixes


class BlackListValidator:
    """
    指定黑名单并进行验证，
    可以用于 global_validator
    """

    def __init__(self, blacklist):
        self._excluded_entries = None
        if not blacklist:
            return

        if isinstance(blacklist, Iterable):
            self._excluded_entries = (Path(str(x)).absolute() for x in blacklist)
        else:
            self._excluded_entries = Path(str(blacklist))

    def __call__(self, entry: Path):
        if not self._excluded_entries:
            return True
        return entry.absolute() in self._excluded_entries


class MultiValidator:
    """融合多个 validator"""

    def __init__(self, *validators):
        self._validators = set()
        for val in validators:
            self.add(val)

    def add(self, validator: Callable[[Path], bool]):
        if isinstance(validator, MultiValidator):
            self._validators = self._validators.union(validator._validators)
        else:
            self._validators.add(validator)
        return self

    def __call__(self, entry: Path) -> bool:
        if not self._validators:
            return True
        for val in self._validators:
            if not val(entry):
                return False
        return True



from pathlib import Path
from cx_core.text.text_utils import quote_text
from .env import env
from .mission_maker import Mission


class ScriptWriter:
    def __init__(self, target) -> None:
        self.target = Path(target).absolute()
        self.output = None
        self.planned_folders = set()
        self.task = None

    def __enter__(self):
        self.task = env.progress.add_task(description="写入脚本文件…", total=None)
        self.output = open(self.target, "wt")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        result = False
        self.output.flush()
        self.output.close()
        env.print(f"脚本文件 [cyan]{self.target.absolute()}[/cyan] 已关闭")
        env.progress.stop_task(self.task)
        env.progress.remove_task(self.task)
        if exc_type is None:
            pass
        return result

    @staticmethod
    def compile_cmd(mission: Mission):
        line = [mission.ffmpeg]
        line += mission.general.iter_arguments()
        for i in mission.inputs:
            line += i.iter_arguments()
            line += ["-i", quote_text(i.filename)]
        for o in mission.outputs:
            line += o.iter_arguments()
            line.append(quote_text(o.filename))
        return " ".join([str(item) for item in line if item is not None])

    def write(self, mission: Mission):
        env.progress.update(
            self.task,
            description=f"正在为[cyan]{mission.source.name}[/cyan]检测目标目录…",
        )
        for o in mission.outputs:
            folder = o.filename.parent
            if folder not in self.planned_folders:
                self.output.write(f"mkdir -p {quote_text(folder.absolute())}\n")
                self.planned_folders.add(folder)
                env.debug(f"添加新建目录指令:{folder}")

        env.progress.update(
            self.task, description=f"为[cyan]{mission.source.name}[/cyan]写入命令…"
        )
        cmd = ScriptWriter.compile_cmd(mission)
        self.output.write(f"{cmd}\n")
        env.info(f"写入命令 [purple]{cmd}[/purple]")

    def write_all(self, missions: list[Mission]):
        for m in env.progress.track(missions, task_id=self.task):
            self.write(m)



class SourceFileError(IOError):
    pass


class TargetFileError(IOError):
    pass


class NoLoaderError(SourceFileError):
    pass


class TargetConflictError(TargetFileError):
    pass


class TargetExistsError(TargetFileError):
    pass


class NoSaverError(TargetFileError):
    pass


class UserCancelError(Exception):
    pass



from cx_core.app import AbstractApp, LogLevel
from cx_core.misc import DataPackage
from argparse import ArgumentParser

from rich.markdown import Markdown
from rich.panel import Panel
from rich_argparse import RichHelpFormatter
from pathlib import Path
from .env import env

from .profile_loader import ProfileLoader
from .source_detector import SourceDetector

from .exceptions import *
import pkgutil
from cx_core.filesystem.path_utils import *
from .mission_maker import MissionMaker, Mission
from .script_writer import ScriptWriter
from .transcoder import Transcoder
from cx_core.tui import JobCounter
from ffmpeg import FFmpegError
import itertools


class MediaKillerApp(AbstractApp):
    APP_VERSION = "0.4.3.2"
    APP_NAME = "mediakiller"

    def __init__(self):
        super(MediaKillerApp, self).__init__()

        _parser = ArgumentParser(
            prog=MediaKillerApp.APP_NAME,
            formatter_class=RichHelpFormatter,
            description="批量转码工具",
            epilog=f"Version {MediaKillerApp.APP_VERSION} Designed by xiii_1991",
        )

        _parser.add_argument(
            "sources",
            help="指定需要处理的文件，其中必须包含至少一个配置文件",
            default=None,
            metavar="需要处理的路径",
            nargs="*",
        )
        _parser.add_argument(
            "-g",
            "--generate",
            action="store_true",
            dest="generate_example",
            help="生成范例配置文件",
        )
        _parser.add_argument(
            "-s",
            "--make-script",
            dest="script_output",
            metavar="脚本文件路径",
            help="生成对应的脚本文件",
        )
        _parser.add_argument(
            "-o",
            "--output",
            dest="output_dir",
            metavar="输出位置",
            help="指定输出目录",
            default=".",
        )
        _parser.add_argument(
            "-c",
            "--continue",
            dest="continue_mode",
            action="store_true",
            help="继续未完成的任务",
        )
        _parser.add_argument(
            "--no-sort", dest="no_sort", action="store_true", help="禁用任务自动排序"
        )
        _parser.add_argument(
            "--pretend",
            "-p",
            dest="pretend_mode",
            action="store_true",
            help="空转模式，模拟执行",
        )
        _parser.add_argument(
            "-d", "--debug", action="store_true", dest="debug", help="显示调试信息"
        )
        _parser.add_argument(
            "--full-help", help="显示详细的说明", dest="full_help", action="store_true"
        )
        _parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=MediaKillerApp.APP_VERSION,
            help="显示软件版本信息",
        )

        self.parser = _parser
        self.profile = None
        self.args = None
        self.global_task = None

    def __enter__(self):
        env.start()
        env.print(
            f"[yellow]{MediaKillerApp.APP_NAME}[/yellow] [blue]{MediaKillerApp.APP_VERSION}[/blue]"
        )

        _args = self.parser.parse_args()
        self.args = DataPackage(**vars(_args))
        env.log_level = LogLevel.DEBUG if self.args.debug else LogLevel.WARNING
        env.debug("解析命令行参数：", self.args)

        self.global_task = env.progress.add_task("全局进度", start=False, visible=False)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        env.progress.stop_task(self.global_task)
        env.progress.remove_task(self.global_task)
        self.global_task = None

        result = False
        if exc_type is None:
            pass
        elif issubclass(exc_type, CxException):
            env.error(exc_val)
            result = True

        env.stop()
        return result

    @staticmethod
    def copy_example_profile(to):
        to = normalize_path(to)
        to = force_suffix(to, ".toml")
        data = pkgutil.get_data("media_killer", "example_project.toml")
        try:
            with open(to, "xb") as file:
                file.write(data)
            env.print(f"配置文件{to.name}已初始化，[red]请在修改后运行[/red]")
        except FileExistsError:
            env.error(
                f"文件 [yellow]{to.name}[/yellow] 已存在，请手动删除或指定其它目标文件"
            )

    @staticmethod
    def show_full_help():
        data = pkgutil.get_data("media_killer", "help.md").decode("utf_8")
        panel = Panel(Markdown(data), width=80)
        env.console.print(panel)

    def _load_profiles(self) -> list[DataPackage]:
        env.progress.update(self.global_task, description="检测配置文件…")
        profiles = []
        pro_files = list(filter(lambda x: Path(x).suffix == ".toml", self.args.sources))
        env.print(f"在输入中发现[green]{len(pro_files)}[/green]个配置文件")
        for a in pro_files:
            with ProfileLoader(a, self.args) as loader:
                profiles.append(loader.load())
        if len(profiles) == 0:
            raise NoProfileError("没有指定配置文件，无法执行任务")
        return profiles

    def _detect_sources(self) -> list[Path]:
        env.progress.update(self.global_task, description="探测源文件…")
        result = []
        sources = set(filter(lambda x: Path(x).suffix != ".toml", self.args.sources))
        source_count = len(sources)
        env.print(f"发现[cyan]{source_count}[/cyan]个来源路径")
        with SourceDetector(self.args) as detector:
            for s in sources:
                detector.detect(s)
            result = detector.arrange_tasks()
        result_count = len(result)
        if result_count != source_count:
            env.print(f"展开后探测到[cyan]{result_count}[/cyan]个源文件")
        return result

    def _make_missions(
        self, profiles: list[DataPackage], sources: list[Path]
    ) -> list[Mission]:
        missions = []
        env.progress.update(
            self.global_task,
            description="制定计划中…",
            completed=0,
            total=len(profiles) * len(sources),
        )
        for profile, source in itertools.product(profiles, sources):
            maker = MissionMaker(source, profile)
            mission = maker.make()
            missions.append(mission)
            env.progress.advance(self.global_task)
        total_count = len(missions)
        env.print(
            f"为[green]{len(profiles)}[/green]个配置文件生成了[yellow]{total_count}[/yellow]个任务"
        )

        if self.args.continue_mode:
            env.debug("检测到继续模式，开始检查已完成的任务")
            env.progress.update(
                self.global_task, description="检测已完成的任务…", total=None
            )
            filtered_missions = [
                m
                for m in missions
                if not all([x.exists() for x in m.iter_output_filenames()])
            ]
            env.print(
                f"已跳过[red]{total_count - len(filtered_missions)}[/red]项已完成的任务"
            )
            missions = filtered_missions

        if not self.args.no_sort:
            env.progress.update(self.global_task, description="正在按路径对任务排序…")
            missions.sort(key=lambda x: str(x.source.resolve()))

        return missions

    def _pretend_run_missions(self, missions: list[Mission]):
        mission_count = len(missions)
        env.progress.update(
            self.global_task, description="假装执行任务…", total=mission_count
        )
        job_counter = JobCounter(mission_count)
        for m in missions:
            if env.wanna_quit:
                raise UserCanceledError("用户取消了执行")
            job_counter.advance()
            filenames = [str(x.filename) for x in m.outputs]
            for name in filenames:
                env.print(
                    f"[yellow]{job_counter}[/yellow] 假装生成了 [cyan]{name}[/cyan]"
                )
            env.progress.advance(self.global_task)

    def _export_script(self, missions: list[Mission], script: Path):
        script = Path(script)
        env.progress.update(self.global_task, description="生成脚本文件…", total=None)
        with ScriptWriter(script) as writer:
            writer.write_all(missions)

    def run(self):
        if self.args.full_help:
            env.info("检测到full_help，打印帮助文件并输出")
            MediaKillerApp.show_full_help()
            return

        if not self.args.sources:
            env.info("指定的路径为空")
            raise NoSourcesError("未指定任何文件，你想做什么？")

        if self.args.generate_example:
            MediaKillerApp.copy_example_profile(self.args.sources[0])
            return

        env.progress.update(self.global_task, visible=True, total=None)
        env.progress.start_task(self.global_task)

        profiles = self._load_profiles()
        total_sources = self._detect_sources()
        missions = self._make_missions(profiles, total_sources)
        mission_count = len(missions)

        env.progress.start_task(self.global_task)
        env.progress.update(self.global_task, completed=0, total=mission_count)
        job_counter = JobCounter(mission_count)
        env.progress.update(
            self.global_task,
            description="总体进度",
        )

        if self.args.pretend_mode:
            env.print("启用了干转模式，将会假装执行")
            self._pretend_run_missions(missions)
            return

        if self.args.script_output:
            env.debug(
                f"指定了脚本目标[cyan]{script_target.name}[/cyan]，将会生成脚本文件"
            )
            self._export_script(missions, self.args.script_output)

        else:
            for m in missions:
                mission_name = m.source.name
                if env.wanna_quit:
                    env.print("取消未完成的任务…")
                    raise UserCanceledError("用户取消了执行")
                skipped = False
                try:
                    with Transcoder(m) as coder:
                        coder.run()
                except FFmpegError as e:
                    env.error(f"FFMPEG运行出错 ：[red]{e.message}[/red]")
                    skipped = True

                job_counter.advance()
                env.progress.update(
                    self.global_task,
                    completed=job_counter.current,
                    total=job_counter.total,
                )
                if skipped:
                    env.print(
                        f"[yellow]{job_counter}[/yellow] [red]{mission_name}[/red] 未正确执行"
                    )
                else:
                    env.print(
                        f"[yellow]{job_counter}[/yellow] [cyan]{mission_name}[/cyan] 执行完毕"
                    )


def run():
    with MediaKillerApp() as media_killer:
        media_killer.run()



from pathlib import Path


def normalize_path(path: Path, anchor=Path(".")) -> Path:
    path = Path(path)
    anchor = Path(anchor)
    if path.is_absolute():
        return path.absolute()

    t = str(path)
    if t.startswith("~"):
        return Path(t.replace("~", str(Path.home()))).absolute()

    return (anchor / path).resolve(strict=False)


def normalize_suffix(suffix: str, with_dot=True) -> str:
    """检查扩展名是否带点"""
    s = str(suffix).strip().strip(".").lower()
    return "." + s if with_dot else s


def force_suffix(source: Path, suffix: str, ignore_dir=True):
    if not source:
        return None
    source = Path(source)
    if source.is_dir() and ignore_dir:
        return source
    suffix = normalize_suffix(suffix)
    return source if source.suffix == suffix else source.with_suffix(suffix)


def quote_path(path, quote_char='"') -> str:
    path = str(path)
    return f"{quote_char}{path}{quote_char}" if " " in path else path


def is_file_in_dir(file, dir) -> bool:
    f = str(Path(file).resolve().absolute())
    d = str(Path(dir).resolve().absolute())
    return f in d



import os
import platform
from functools import cached_property
from pathlib import Path


class OSInfo:
    def __init__(self):
        self.__system = platform.system()

    @cached_property
    def system(self):
        return self.__system

    @cached_property
    def hosts_file(self):
        if self.system == 'Windows':
            win_dir = Path(os.getenv('windir'))
            return win_dir / "System32/drivers/etc/hosts"
        return Path("/etc/hosts")



from cx_core import DataPackage
from cx_subtitle import SubtitleFormatter
from cx_subtitle.loader import *
from cx_subtitle.saver import *
from .env import env
from .exceptions import *
from .subtitle_translator import SubtitleTranslator


class Mission:
    def __init__(self, source, profile: DataPackage):
        self._profile = profile
        self._source = Path(source).absolute()

    @property
    def profile(self):
        return self._profile

    @property
    def source(self) -> Path:
        return self._source

    @property
    def name(self) -> str:
        return self._source.name

    @cached_property
    def target_dir(self) -> Path:
        t_dir = self._source.parent
        if self.profile.target_dir:
            t_dir = Path(self.profile.target_dir).absolute()
        if self.profile.sub_dir:
            t_dir = t_dir / str(self.profile.sub_dir)
        return t_dir

    @cached_property
    def target_basename(self) -> str:
        result = self._source.stem
        if self.profile.appending:
            result += str(self.profile.appending)
        return result

    def make_target(self, suffix: str):
        if not suffix.startswith('.'):
            suffix = '.' + suffix
        return self.target_dir / (self.target_basename + suffix)


class Worker:
    __loader_classes = {
        '.srt': SrtLoader,
        '.txt': TxtLoader,
        '.docx': WordLoader,
        '.ttml': TTMLLoader
    }

    __saver_classes = {
        'srt': SrtSaver,
        'txt': TxtSaver,
        'word': WordSaver,
        'excel': ExcelSaver,
        'csv': CsvSaver
    }

    __suffixes = {
        'txt': '.txt',
        'word': '.docx',
        'excel': '.xlsx',
        'csv': '.csv',
        'srt': '.srt',
        'ttml': '.ttml',
        'rtf': '.rtf'
    }

    _formatter = None
    _translator = None

    def __init__(self, mission: Mission):
        self.mission = mission
        self.task = None

    @staticmethod
    def acceptable_suffixes():
        return Worker.__loader_classes.keys()

    @staticmethod
    def supported_formats():
        return Worker.__saver_classes.keys()

    @property
    def profile(self):
        return self.mission.profile

    @cached_property
    def target_suffix(self):
        res = Worker.__suffixes.get(self.profile.format)
        return res or '.txt'

    @cached_property
    def source_suffix(self):
        return self.mission.source.suffix

    @cached_property
    def loader_class(self):
        return Worker.__loader_classes.get(self.source_suffix)

    @cached_property
    def saver_class(self):
        return Worker.__saver_classes.get(self.profile.format)

    @cached_property
    def target(self):
        return self.mission.make_target(self.target_suffix)

    def __enter__(self):
        if not self.profile.bypass_formatter and not Worker._formatter:
            env.print(f'启用[blue]字幕格式检查器')
            Worker._formatter = SubtitleFormatter()

        if self.profile.translate and not Worker._translator:
            env.print(f'启用[blue]字幕文本翻译器')
            Worker._translator = SubtitleTranslator()

        self.task = env.progress.add_task(self.mission.name, total=None)
        env.debug(f'Worker启动: [yellow]{self.mission.source}[/yellow] --> [cyan]{self.target}[/cyan]')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        env.progress.remove_task(self.task)
        self.task = None
        env.debug(f'删除 Worker 的进度条')

        result = False
        if exc_type is None:
            env.print(f'[cyan]{self.target.name}[/cyan]已完成')
        elif issubclass(exc_type, NoLoaderError):
            env.warning(f'未为来源文件 [yellow]{self.mission.name}[/yellow] 找到合适的加载器，'
                        f'[red]将跳过此任务[/red]')
            result = True
        elif issubclass(exc_type, NoSaverError):
            env.warning(f'未为目标文件 [cyan]{self.target.name}[/cyan] 找到合适的加载器，'
                        f'[red]将跳过此任务[/red]')
            result = True
        elif issubclass(exc_type, TargetConflictError):
            env.warning(
                f'来源文件 [yellow]{self.mission.name}[/yellow] 和目标文件 [cyan]{self.target.name}[/cyan] '
                f'的[red]路径与文件名相同,将跳过此任务[/red]')
            result = True
        elif issubclass(exc_type, TargetExistsError):
            env.warning(
                f'目标文件 [cyan]{self.target.name}[/cyan] 已经存在，'
                f'[red]请更改目标位置或手动移除此文件[/red]，本次将跳过此任务')
            result = True
        elif issubclass(exc_type, UserCancelError):
            env.error(
                f'[red]尝试移除未完成的目标文件: [cyan]{self.target.name}[/cyan][/red]')
            self.target.unlink(missing_ok=True)
            result = True

        return result

    def process(self):
        if not self.loader_class:
            raise NoLoaderError()

        if not self.saver_class:
            raise NoSaverError()

        if self.mission.source == self.target:
            raise TargetConflictError()

        if self.target.exists():
            raise TargetExistsError()

        with self.loader_class(self.mission.source,
                               encoding=self.profile.source_encoding) as loader:
            env.info('构建loader...')
            if not self.target.parent.exists():
                env.print(f'目标目录[blue]{self.target.parent.absolute()}[/blue]不存在，正在创建...')
                self.target.parent.mkdir(parents=True, exist_ok=True)

            with self.saver_class(self.target,
                                  keep_time=self.profile.keep_time,
                                  encoding=self.profile.target_encoding) as saver:
                env.info('构建saver...')
                saver.install_processor(Worker._translator) \
                    .install_processor(Worker._formatter)

                env.debug('开始读写...')
                for subtitle in env.progress.track(loader.subtitles(), task_id=self.task):
                    saver.write(subtitle)
                    env.debug(f'写入文本: [yellow]{subtitle.content}[/yellow]')
                    env.progress.update(self.task, description=f'[yellow]{subtitle.content}[/yellow]')
                    if env.wanna_quit:
                        raise UserCancelError

        env.debug(f'读写操作完毕 {self.target}')



import time

import translators
from xpinyin import Pinyin as pinYin

from cx_subtitle import Subtitle, SubtitleProcessor
from .env import env


class SubtitleTranslator(SubtitleProcessor):
    MAX_SEQUENCE = 120
    SLEEP_SECOND = 15

    def __init__(self, target_lang='en'):
        super(SubtitleTranslator, self).__init__()
        env.debug('初始化翻译器')
        self.target_lang = target_lang
        self.ts = translators
        self._count = 0
        self._pin_yin = None

    def _fallback_translate(self, s: str):
        if not self._pin_yin:
            self._pin_yin = pinYin()
        return self._pin_yin.get_pinyin(s, ' ')

    def __call__(self, s: Subtitle):
        translated_content = s.content
        for i in range(10):
            if env.wanna_quit:
                break
            try:
                translated_content = str(self.ts.translate_text(s.content, to_language=self.target_lang))
                env.debug(f'翻译文本 "{s.content}" 为 "{translated_content}" ')
                break
            except KeyboardInterrupt:
                env.error(f'用户终止服务')
                break
            except Exception as e:
                env.error(f'ERROR {i}: [red]{e}[/red]')
                time.sleep(2)
        if translated_content == s.content:
            env.debug(f'[yellow]{translated_content}[/yellow]，在线翻译失败，将启动备用方案')
            translated_content = self._fallback_translate(translated_content)

        self._count += 1
        if self._count >= SubtitleTranslator.MAX_SEQUENCE:
            env.info('[cyan]达到连续最大请求数，正在等待……[/cyan]')
            time.sleep(SubtitleTranslator.SLEEP_SECOND)
            self._count = 0

        return s.with_content(translated_content)



import os
import urllib.request
from argparse import ArgumentParser

from rich_argparse import RichHelpFormatter

from cx_core import DataPackage
from cx_core.app import OSInfo, AbstractApp
from .env import env


class GithubHostsUpdaterApp(AbstractApp):
    URL = "https://gitlab.com/ineo6/hosts/-/raw/master/next-hosts"

    APP_NAME = 'update_githubhosts'
    APP_VERSION = '0.1.10'

    def __init__(self):
        super(GithubHostsUpdaterApp, self).__init__()
        self.osinfo = OSInfo()
        parser = ArgumentParser(prog=GithubHostsUpdaterApp.APP_NAME, formatter_class=RichHelpFormatter,
                                description='自动更新githubhosts', epilog='Designed by xiii_1991')
        parser.add_argument('-p', '--pretend', action='store_true', dest='pretend',
                            help='模拟执行并输出结果')
        parser.add_argument('-s', '--save', action='append', dest='side_save', metavar='PATH',
                            help='输出到指定位置')
        parser.add_argument('--version', action='version', version=GithubHostsUpdaterApp.APP_VERSION,
                            help="显示版本号")
        self._parser = parser
        self.status = None
        self.args = None

    def refresh_dns(self):
        self.status.update('尝试刷新DNS...')
        if self.osinfo.system == 'Darwin':
            os.system("killall -HUP mDNSResponder")
            env.print('杀死 mDNSResponder')
        elif self.osinfo.system == 'Linux':
            os.system("systemctl restart systemd-resolved")
            env.print('重启网络服务')
        elif self.osinfo.system == 'Windows':
            os.system("ipconfig /flushdns")
            env.print('刷新 DNS 信息')
        else:
            env.warning('无法识别操作系统，请手动刷新')

    def __enter__(self):
        self.status = env.console.status('正在启动...')
        env.print(f'[yellow]{GithubHostsUpdaterApp.APP_NAME}[/yellow] [blue]{GithubHostsUpdaterApp.APP_VERSION}[/blue]')
        self.args = DataPackage(**vars(self._parser.parse_args()))
        self.status.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.status.update('执行完毕')
        self.status.stop()
        return False

    def run(self):
        self.status.update('正在下载新的hosts...')
        local_filename, headers = urllib.request.urlretrieve(self.URL)
        env.print("已下载新的 Github Hosts")

        self.status.update('开始解析...')
        github_hosts = []
        with open(local_filename, encoding='utf-8') as new_hosts:
            for line in new_hosts:
                github_hosts.append(line)

        old_hosts = []
        with open(self.osinfo.hosts_file, encoding='utf-8') as host_file:
            inside_github_section = False
            for line in host_file:
                if line == github_hosts[0]:
                    inside_github_section = True
                    continue
                if line == github_hosts[-1]:
                    inside_github_section = False
                    continue
                if not inside_github_section:
                    old_hosts.append(line)

        env.print('解析完毕')

        if self.args.pretend:
            self.status.update('正在预览结果...')
            env.print(''.join(old_hosts + github_hosts))
            return

        self.status.update('正在写入...')
        target = self.osinfo.hosts_file
        if self.args.side_save:
            target = self.args.side_save[0]
            env.print(f'修改输出目标为 {target}')

        with open(target, "w", encoding='utf-8') as new_host:
            new_host.writelines(old_hosts)
            new_host.writelines(github_hosts)
            env.print('写入完毕。')

        self.refresh_dns()


def run():
    with GithubHostsUpdaterApp() as app:
        app.run()



from cx_core.filesystem import PathExpander, SuffixValidator
from pathlib import Path
from typing import Iterable
from .env import env
from . import source_adapter as Adapters
import concurrent.futures


class SourceDetector:
    DEFAULT_SUFFIXES = {
        ".mov",
        ".mkv",
        ".mp4",
        ".flv",
        ".3gp",
        ".3gpp",
        ".rmvb",
        ".mp3",
        ".aac",
        ".mxf",
        ".mxf_op1a",
        ".vob",
        ".wmv",
        ".wma",
        ".srt",
        ".ass",
        ".aas",
        ".ttml",
        ".ogg",
        ".oga",
        ".ogv",
        ".m4a",
        ".m4v",
        ".3g2",
        ".mpeg",
        ".mpg",
        ".ts",
        ".lrc",
        ".h264",
        ".flac",
        ".ast",
        ".asf",
        ".gif",
    }

    ADDITIONAL_SUFFIXES = {".xml", ".txt", "csv"}

    def __init__(self, args=None) -> None:
        self.args = args
        self.task = None
        self.exp_settings = None
        self._sources = set()

    def __enter__(self):
        self.task = env.progress.add_task("解析源路径")
        suffixes = SourceDetector.DEFAULT_SUFFIXES
        if self.args.source:
            includes = {
                "." + str(x).strip().strip(".")
                for x in self.args.source.suffix_includes
            }
            excludes = {
                "." + str(x).strip().strip(".")
                for x in self.args.source.suffix_excludes
            }
            if self.args.ignore_default_suffixes:
                suffixes.clear()
            suffixes = suffixes | includes - excludes
        self.acceptable_suffixes = suffixes
        env.debug("最终的可接受扩展名白名单为：", self.acceptable_suffixes)
        self.exp_settings = PathExpander.Settings(
            file_validator=SuffixValidator(self.acceptable_suffixes), accept_dir=False
        )
        return self

    def __exit__(self, et, ev, eb):
        env.progress.remove_task(self.task)
        return False

    def detect(self, source):
        source = Path(source)
        if not source.exists():
            return

        new_tasks = []

        env.progress.update(self.task, description="正在探测文件…")

        if source.is_dir():
            expander = PathExpander(source, self.exp_settings)
            new_tasks += [x for x in expander]
            env.print(f"探测到[cyan]{len(new_tasks)}[/cyan]个有效的文件")
        else:
            suffix = source.suffix.lower().strip(".")
            if suffix in Adapters.adapters:
                with Adapters.adapters[suffix](source) as adapter:
                    new_tasks += adapter.items()
            else:
                new_tasks.append(source)

        env.progress.update(self.task, completed=0, total=len(new_tasks))
        for t in new_tasks:
            tt = Path(t)
            self._sources.add(tt)
            env.progress.update(self.task, description=tt.name)
            env.progress.advance(self.task)

    def arrange_tasks(self):
        def filter_task(task):
            a = Path(task)
            if a.exists and a.is_file:
                suffix = a.suffix.lower()
                if suffix in self.acceptable_suffixes:
                    return a
            return None

        env.progress.update(self.task, description="正在整理任务列表…")
        with concurrent.futures.ThreadPoolExecutor() as executor:
            _checked_tasks = executor.map(filter_task, self._sources)
            filtered_tasks = {x for x in _checked_tasks if x is not None}

        env.print(f"整理完成后，共有[cyan]{len(filtered_tasks)}[/cyan]个待处理文件")
        self._sources = filtered_tasks
        return self._sources



import signal

from cx_core.app import ProgressEnv

env = ProgressEnv()


def signal_handler(sig, frame):
    if sig != signal.SIGINT:
        return
    env.info('接收到 SIGINT')
    env.print("收到终止信号，准备退出...")
    env.wanna_quit = True


signal.signal(signal.SIGINT, signal_handler)



import signal

from rich.progress import (
    Progress,
    SpinnerColumn,
    TextColumn,
    BarColumn,
    TaskProgressColumn,
    TimeRemainingColumn,
)
from rich.table import Column

from cx_core.app import AppLogger, AbstractEnv, LogLevel


class Env(AbstractEnv):
    def __init__(self):
        super(Env, self).__init__()
        self._progress = Progress(
            SpinnerColumn(),
            TextColumn(
                "[progress.description]{task.description}",
                table_column=Column(ratio=50),
            ),
            BarColumn(table_column=Column(ratio=20)),
            TaskProgressColumn(),
            TimeRemainingColumn(),
            expand=True,
            transient=True,
        )

        self._logger = AppLogger(console=self._progress.console)
        self._logger.level = LogLevel.WARNING

        self.args = None

        self.wanna_quit = False

    def start(self):
        self._progress.start()
        self.info("全局环境已启动")

    def stop(self):
        self._progress.stop()
        self.info("Bye~")

    @property
    def progress(self):
        return self._progress

    @property
    def console(self):
        return self._progress.console


env = Env()


def signal_handler(sig, frame):
    if sig != signal.SIGINT:
        return
    env.info("接收到 SIGINT")
    env.print("[red]收到终止信号，准备退出...[/red]")
    env.wanna_quit = True


signal.signal(signal.SIGINT, signal_handler)
